# wdgeomap.tcl --
#
#	This file provides the tkgeomap_procs package, which augments the
#	tkgeolinearray and tkgeoplace extensions.  See the tkgeomap_procs (n)
#	man page for details.
# 
# Copyright (c) 2005 Gordon D. Carrie. All rights reserved.
# 
# Licensed under the Open Software License version 2.1
# 
# Please address questions and feedback to user0@tkgeomap.org
# 
#  @(#) $Id: wdgeomap.tcl,v 1.47 2008/08/26 21:50:54 tkgeomap Exp $

package require tkgeomap 2
package require tclgeomap_procs 2
package provide wdgeomap 2.11.5

namespace eval ::geomap::wdgeomap {}

# ::geomap::wdgeomap::create --
#
#	This procedure creates an interactive geographic
#
# Arguments:
#	map		- map identifier, name of map command.
#	map_frame	- frame to hold map and controls.
#	args		- configuration arguments.
#
# Results:
#	See the user documentation.

proc ::geomap::wdgeomap::create {map map_frame args} {

    # Create widgets and initialize a private namespace for the map.

    set ns ::geomap::wdgeomap::$map
    set ns [namespace eval $ns {namespace current}]
    variable map_ns
    set map_ns($map) $ns
    frame $map_frame
    set ${ns}::map_frame $map_frame
    variable map_tl
    set toplev [winfo toplevel $map_frame]
    set map_tl($toplev) $map
    set map_canvas ${map_frame}.canvas
    canvas $map_canvas
    set ${ns}::map_canvas $map_canvas

    # This list determines the stacking order for map items.

    set ${ns}::layers {}

    # These bindtags will be used to adjust motion bindings as map position
    # and projection change.

    bindtags $map_canvas \
	    [linsert [bindtags $map_canvas] 0 geomap_center geomap_motion]

    # These are the canvas coordinates of the map center.
    # Initialize them with bogus values, then set them properly with the
    # setcenter procedure, which is defined below.

    set ${ns}::xCtr 0
    set ${ns}::yCtr 0
    setcenter $map
    bind geomap_center <Configure> [namespace code [list setcenter $map]]

    # Initialize map parameters.

    variable ${ns}::refpoint [list 0.0 0.0]
    variable ${ns}::projname Mercator
    variable ${ns}::projId [::geomap::projection Mercator 0.0]
    variable ${ns}::scale 1.0e-7
    variable ${ns}::boundcirclecolor Black
    variable ${ns}::update ""
    variable ${ns}::bindings {}
    variable ${ns}::lazy 0
    variable ${ns}::projections {CylEqDist Mercator CylEqArea LambertConfConic \
	    LambertEqArea Stereographic PolarStereographic Orthographic}
    variable ${ns}::scales {1:10000000 1:20000000 1:30000000 1:45000000 \
	    1:60000000 1:90000000 1:120000000}
    variable ${ns}::rotationmenu 1

    # Create the menu bar and default menus.

    set ${ns}::mbar [frame ${map_frame}.mbar -relief raised -borderwidth 3]
    set m [addmenu $map "File"]
    $m add command -label "Postscript" \
	    -command [namespace code [list postscript $map]]
    $m add command -label "Quit" -command {exit}
    addmenu $map "Projection"
    addmenu $map "Scale"
    set m [addmenu $map "Top"]
    foreach r {north ne east se south sw west nw} {
	set cmd [namespace code [list configure $map -rotation $r]]
	$m add command	-label $r -command $cmd
    }
    unset -nocomplain r

    # Pack widgets in main frame.
    # Frame should be positioned by caller.

    pack forget $map_canvas
    pack ${map_frame}.mbar -fill x
    pack $map_canvas -fill both -expand true

    # Create a bogus item.  This ensures that calculations that
    # assume the existence of at least one item with the geomap tag
    # will always work.

    draw $map geomap_place ""

    # Initialize configuration

    configure $map -projname $projname -projections $projections \
	    -scales $scales -rotationmenu $rotationmenu
    if {[llength $args] > 0} {
	set cmd [linsert $args 0 configure $map]
	eval $cmd
    }
}

# ::geomap::wdgeomap::map_proc --
#
#	This is the callback for map commands created by the wdgeomap
#	procedure.
#
# Arguments:
#	map	- map identifier.
#	cmd	- command to execute.
#	args	- options to the command.
#
# Results:
#	A procedure corresponding to cmd is called with map as one of
#	its arguments.

proc ::geomap::wdgeomap::map_proc {map cmd args} {
    set commands [list							\
	"create" "map_proc" "name" "canvas" "mbar" "color_dialog"	\
	"xytolatlon" "latlontoxy"					\
	"setcenter" "configure" "cget" "draw" "erase"			\
	"set_layers" "get_layers" "add_layer" "rm_layer"		\
	"layers_dlg_drag" "layers_dlg_brelease" "set_layers_dlg" 	\
	"item_cmp" "y_cart" "y_canvas"					\
	"addmenu" "getmenu" "deletemenu"				\
	"ctr_reflon" "ctr_refpt"					\
	"setcolor" "setcolorscript" "getcolor"				\
	"choose_color" "load_colors" "save_colors" "undo_colors"	\
	"set_colors_dlg" "postscript" "delete"]
    if {[lsearch -exact $commands $cmd] < 0} {
	error "$map: unknown subcommand \"$cmd\""
    }
    eval [linsert $args 0 $cmd $map]
}

# ::geomap::wdgeomap::name --
#
#	Create a unique map name.  This should be used in window titles,
#	widget names, and other places that need an identifier that will
#	not conflict with other map names.
#
# Arguments:
#	map	- map identifier
#
# Results:
#	Return value is fully qualified map name, with the :: separators
#	replaced with _ characters.

proc ::geomap::wdgeomap::name {map} {
    return [string map {:: _} [string range $map 2 end]]
}

# ::geomap::wdgeomap::map_canvas --
#
#	Return the name of the map_canvas.
#
# Arguments:
#	map	- map identifier

proc ::geomap::wdgeomap::map_canvas {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    return $map_canvas
}

# ::geomap::wdgeomap::mbar --
#
#	Return the name of the menu bar.
#
# Arguments:
#	map	- map identifier

proc ::geomap::wdgeomap::mbar {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::mbar
    return $mbar
}

# ::geomap::wdgeomap::xytolatlon --
#
#	Convert x y canvas coordinates in
#	the map canvas to {latitude longitude}
#
# Arguments:
#	map	- map identifier.
#	args	- list of canvas coordinates of form
#		  {?-catch? x1 y1 x2 y2 ...}
#		  If first word of args is "-catch" skip points that
#		  cannot be converted.
#
# Results:
#	Return value is a list of form {{lat1 lon1} {lat2 lon2} ...}
#	giving the geographic coordinates of the canvas points.

proc ::geomap::wdgeomap::xytolatlon {map args} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::refpoint
    variable ${ns}::projId
    variable ${ns}::scale

    if {[lindex $args 0] == "-catch"} {
	set catch 1
	set args [lrange $args 1 end]
    } else {
	set catch 0
    }

    # Map coordinates of reference point

    set mpt0 [::geomap::scalept [$projId fmlatlon $refpoint] $scale]
    set abs0 [lindex $mpt0 0]
    set ord0 [lindex $mpt0 1]

    # Map coordinates of args = {x y ...}

    set coords [$map_canvas coords geomap]
    set x0 [lindex $coords 0]
    set y0 [lindex $coords 1]
    set mPerPx [expr {0.0254 / 72.0 / [tk scaling]}]
    set l [llength $args]
    set s_inv [expr {1.0 / $scale}]
    if {$l == 2} {
	set x [lindex $args 0]
	set y [lindex $args 1]
	set abs [expr {$abs0 - ($x0 - $x) * $mPerPx}]
	set ord [expr {$ord0 + ($y0 - $y) * $mPerPx}]
	set mpt [list $abs $ord]
	set mpt [::geomap::scalept $mpt $s_inv]
	if $catch {
	    if {[catch [list $projId tolatlon $mpt] gpt] == 0} {
		return $gpt
	    } else {
		return {}
	    }
	} else {
	    return [$projId tolatlon $mpt]
	}
    } elseif {$l % 2 == 0} {
	set r {}
	if $catch {
	    foreach {x y} $args {
		set abs [expr {$abs0 - ($x0 - $x) * $mPerPx}]
		set ord [expr {$ord0 + ($y0 - $y) * $mPerPx}]
		set mpt [list $abs $ord]
		set mpt [::geomap::scalept $mpt $s_inv]
		if {[catch [list $projId tolatlon $mpt] gpt] == 0} {
		    lappend r $gpt
		}
	    }
	} else {
	    foreach {x y} $args {
		set abs [expr {$abs0 - ($x0 - $x) * $mPerPx}]
		set ord [expr {$ord0 + ($y0 - $y) * $mPerPx}]
		set mpt [list $abs $ord]
		set mpt [::geomap::scalept $mpt $s_inv]
		lappend r [$projId tolatlon $mpt]
	    }
	}
	return $r
    } else {
	error "Number of coordinates cannot be odd"
    }
}

# ::geomap::wdgeomap::latlontoxy --
#
#	Convert a lat-lon values to x y
#	coordinates in the map_canvas window.
#
# Arguments:
#	map	- map identifier.
#	args	- a list of form {-catch {lat1 lon1} {lat2 lon2} ...}
#	giving geographic locations.
#
# Results:
#	Return value is a list of form {x1 y1 x2 y2 ...} giving the
#	canvas coordinates corresponding to the input lat-lon's.

proc ::geomap::wdgeomap::latlontoxy {map args} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::refpoint
    variable ${ns}::projId
    variable ${ns}::scale

    if {[lindex $args 0] == "-catch"} {
	set catch 1
	set args [lrange $args 1 end]
    } else {
	set catch 0
    }

    set coords [$map_canvas coords geomap]
    set x0 [lindex $coords 0]
    set y0 [lindex $coords 1]
    set pxPerM [expr {72.0 * [tk scaling] / 0.0254}]

    # Map coordinates of reference point

    set mpt0 [::geomap::scalept [$projId fmlatlon $refpoint] $scale]
    set absC [lindex $mpt0 0]
    set ordC [lindex $mpt0 1]

    # Map coordinates of point

    set r {}
    if $catch {
	foreach geoPt $args {
	    if {[catch [::geomap::scalept [$projId fmlatlon $geoPt] $scale] mpt]
		    == 0} {
		set abs [lindex $mpt 0]
		set ord [lindex $mpt 1]
		set x [expr {$x0 + ($abs - $absC) * $pxPerM}]
		set y [expr {$y0 - ($ord - $ordC) * $pxPerM}]
		lappend r $x $y
	    }
	}
    } else {
	foreach geoPt $args {
	    set mpt [::geomap::scalept [$projId fmlatlon $geoPt] $scale]
	    set abs [lindex $mpt 0]
	    set ord [lindex $mpt 1]
	    set x [expr {$x0 + ($abs - $absC) * $pxPerM}]
	    set y [expr {$y0 - ($ord - $ordC) * $pxPerM}]
	    lappend r $x $y
	}
    }
    return $r
}

# ::geomap::wdgeomap::setcenter --
#
#	Put the map refpoint at the canvas
#	center.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	coords are set for geomap_lnarr and geomap_place items in the
#	canvas.
#	Return value is list of coordinates of map_canvas center.

proc ::geomap::wdgeomap::setcenter {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::xCtr
    variable ${ns}::yCtr

    set xCtr [expr {[winfo width $map_canvas] / 2}]
    set yCtr [expr {[winfo height $map_canvas] / 2}]
    foreach item [$map_canvas find withtag geomap] {
	$map_canvas coords $item $xCtr $yCtr
    }
    return [list $xCtr $yCtr]
}

# ::geomap::wdgeomap::configure --
#
#	Configure map.
#
# Arguments:
#	map	- map identifier.
#	args	- a list of form {-opt arg ?-opt arg ...?} specifying
#		  the configuration options and values.
#
# Results:
#	Namespace variables and the associated widgets are updated.

proc ::geomap::wdgeomap::configure {map args} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_frame
    variable ${ns}::map_canvas
    variable ${ns}::mbar
    variable ${ns}::refpoint
    variable ${ns}::projname
    variable ${ns}::projId
    variable ${ns}::scale
    variable ${ns}::lazy
    variable ${ns}::projections
    variable ${ns}::scales
    variable ${ns}::rotationmenu
    variable ${ns}::bindings
    variable ${ns}::update
    variable ${ns}::boundcirclecolor

    set bound_circle ${ns}::bound_circle
    foreach {opt arg} $args {
	switch -- $opt {
	    -refpoint {
		$map_canvas itemconfigure geomap -refpoint $arg
		set refpoint $arg
		uplevel #0 $update
	    }
	    -projname {
		set projections [::geomap::projections]
		if {[lsearch $projections $arg] < 0} {
		    error "Projection must be one of $projections"
		}
		set projname $arg

		# Delete old bound circle 

		if {[namespace which -command $bound_circle] != ""} {
		    erase $map geomap_lnarr bound_circle
		    rename $bound_circle {}
		}

		# Set projection

		set proj_info [::geomap::proj_info [cget $map -projname]]
		set ref [lindex $proj_info 0]
		set domain [lindex $proj_info 1]
		if {$ref == "longitude"} {
		    set lon [::geomap::longitude $refpoint]
		    eval $projId set $projname $lon
		} elseif {$ref == "point"} {
		    if {$projname == "PolarStereographic"} {
			set lat [::geomap::latitude $refpoint]
			if {$lat > 0.0} {
			    set refpoint {90.0 0.0}
			    eval $projId set PolarStereographic N
			} else {
			    set refpoint {-90.0 0.0}
			    eval $projId set PolarStereographic S
			}
		    } else {
			eval $projId set $projname $refpoint
		    }
		} else {
		    error "Unable to set $projname projection"
		}

		# If necessary, draw new bound circle

		if {$ref == "point" && $domain == "hemisphere"} {
		    ::geomap::lnarr fmlist $bound_circle \
			    [::geomap::circle $refpoint 89.9]
		    draw $map geomap_lnarr $bound_circle	\
			    -outline $boundcirclecolor		\
			    -tags "bound_circle"
		}

		# Run update script

		uplevel #0 $update
	    }
	    -scale {
		if ![string is double $arg] {
		    set scale [::geomap::cartg $arg]
		} else {
		    set scale $arg
		}
		$map_canvas itemconfigure geomap -scale $scale
		uplevel #0 $update
	    }
	    -rotation {
		$projId rotation $arg
		uplevel #0 $update
	    }
	    -boundcirclecolor {
		$map_canvas itemconfigure bound_circle -outline $arg
		set boundcirclecolor $arg
	    }
	    -update {
		set update $arg
	    }
	    -colormenu {
		if ![string is boolean $arg] {
		    error "lazy must be boolean"
		}
		set mb [getmenu $map "File"]
		set m [$mb cget -menu]
		if $arg {
		    if {[catch [list $m index "Colors"]] != 0} {
			set cmd [namespace code [list set_colors_dlg $map]]
			$m insert 0 command -label "Colors" -command $cmd
		    }
		    variable ${ns}::color_dlg .[name $map]_colors
		} else {
		    if {[catch [list $m index "Colors"] i] == 0} {
			$m delete $i
		    }
		}
	    }
	    -layermenu {
		if ![string is boolean $arg] {
		    error "lazy must be boolean"
		}
		set mb [getmenu $map "File"]
		set m [$mb cget -menu]
		if $arg {
		    if {[catch [list $m index "Layers"]] != 0} {
			set cmd [namespace code [list set_layers_dlg $map]]
			$m insert 0 command -label "Layers" -command $cmd
		    }
		} else {
		    if {[catch [list $m index "Layers"] i] == 0} {
			$m delete $i
		    }
		}
	    }
	    -lazy {
		if ![string is boolean $arg] {
		    error "lazy must be boolean"
		}
		set lazy $arg
	    }
	    -projections {
		# Replace the selections in the Projections menu
		# with the given list.

		set projections $arg
		set mb [getmenu $map "Projection"]
		set m [$mb cget -menu]
		$m delete 2 end
		if {[llength $projections] == 0} {
		    pack forget $mb
		} else {
		    foreach p $projections {
			set cmd [namespace code \
				[list configure $map -projname $p]]
			$m add command -label $p -command $cmd
		    }
		    pack $mb -side left
		}
	    }
	    -scales {
		# Replace the selections in the Scales menu
		# with the given list.

		set scales $arg
		set mb [getmenu $map "Scale"]
		set m [$mb cget -menu]
		$m delete 0 end
		if {[llength $scales] == 0} {
		    pack forget $mb
		} else {
		    foreach s $scales {
			set cmd [namespace code [list configure $map -scale $s]]
			$m add command -label $s -command $cmd
		    }
		    pack $mb -side left
		}
	    }
	    -rotationmenu {
		# Display or hide the rotations menu as requested.

		set rotationmenu $arg
		set mb [getmenu $map "Top"]
		if {$rotationmenu} {
		    pack $mb -side left
		} else {
		    pack forget $mb
		}
	    }
	    default {
		if {[catch [list $map_canvas configure $opt $arg] msg] != 0} {
		    error "Could not configure $opt"
		}
	    }
	}
    }
}

# ::geomap::wdgeomap::cget --
#
#	Retrieve a configuration option.
#
# Arguments:
#	map	- map identifier.
#	option	- configuration option whose value is sought.
#
# Results:
#	The value of a configuration option.

proc ::geomap::wdgeomap::cget {map option} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::refpoint
    variable ${ns}::projname
    variable ${ns}::projId
    variable ${ns}::scale
    variable ${ns}::boundcirclecolor
    variable ${ns}::lazy
    variable ${ns}::projections
    variable ${ns}::scales
    variable ${ns}::rotationmenu

    switch -- $option {
	-refpoint {
	    return $refpoint
	}
	-projname {
	    return $projname
	}
	-scale {
	    return $scale
	}
	-rotation {
	    return [$projId rotation]
	}
	-boundcirclecolor {
	    return $boundcirclecolor
	}
	-projname {
	    return $projname
	}
	-lazy {
	    return $lazy
	}
	-projections {
	    return $projections
	}
	-scales {
	    return $scales
	}
	-rotationmenu {
	    return $rotationmenu
	}
	default {
	    if {[catch [list $map_canvas cget $option $arg] msg] != 0} {
		error "Unknown configuration option $option"
	    }
	}
    }
}

# ::geomap::wdgeomap::draw --
#
#	Create a new geomap_lnarr or geomap_place item,
#	or re-configures an old one.
#
# Arguments:
#	map	- map identifier.
#	type	- geomap_lnarr or geomap_place
#	name	- fully qualified name of the lnarr or place.
#	args	- list of form {-option arg ?-option arg ...?} giving
#		  options to pass to the item.  See the tkgeomap (n)
#		  man page for details.
#
# Results:
#	An item is created or configured.
#	Return value is the item identifier for new or modified item.

proc ::geomap::wdgeomap::draw {map type name args} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::xCtr
    variable ${ns}::yCtr
    variable ${ns}::refpoint
    variable ${ns}::projId
    variable ${ns}::scale

    array set config $args
    switch -exact -- $type {
	geomap_lnarr {
	    array set config [list -lnarr $name -refpoint $refpoint	\
		    -projection $projId	-scale $scale]
	    if {$name != ""
		    && [set id [$map_canvas find withtag $name]] != ""
		    && [$map_canvas type $id] == "geomap_lnarr"} {
		# Configure pre-existing geomap_lnarr item.

		eval $map_canvas itemconfigure $id [array get config]
	    } else {
		# Create new geomap_lnarr item.

		set id [eval $map_canvas create geomap_lnarr \
			$xCtr $yCtr [array get config]]
	    }
	}
	geomap_place {
	    array set config [list -place $name -refpoint $refpoint	\
		    -projection $projId	-scale $scale]
	    if {$name != ""
		    && [set id [$map_canvas find withtag $name]] != ""
		    && [$map_canvas type $id] == "geomap_place"} {
		# Configure pre-existing geomap_place item.

		eval $map_canvas itemconfigure $id $args
	    } else {
		# Create new geomap_place item.

		set id [eval $map_canvas create geomap_place \
			$xCtr $yCtr [array get config]]
	    }
	}
	default {
	    error "Type must be \"geomap_lnarr\" or \"geomap_place\""
	}
    }
    $map_canvas addtag $name withtag $id
    $map_canvas addtag geomap withtag $id
    return $id
}

# ::geomap::wdgeomap::erase --
#
#	Deletes and item created with the draw command.
#
# Arguments:
#	map	- map identifier.
#	type	- geomap_lnarr or geomap_place
#	name	- the name of the place or linearray given to the draw
#		  command.
#
# Results:
#	None.

proc ::geomap::wdgeomap::erase {map type name} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas

    if {$type != "geomap_lnarr" && $type != "geomap_place"} {
	error "Type must be \"geomap_lnarr\" or \"geomap_place\""
    }
    foreach id [$map_canvas find withtag $name] {
	if {[$map_canvas type $id] == $type} {
	    $map_canvas delete $id
	    break
	}
    }
}

# ::geomap::wdgeomap::set_layers --
#
#	Sets the drawing sequence for certain canvas items.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	The drawing sequence for items in the canvas may be changed.

proc ::geomap::wdgeomap::set_layers {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::layers

    if { ![info exists layers] } {
	return
    }
    foreach layer $layers {
	if {[llength [$map_canvas find withtag $layer]] > 0} {
	    if { [info exists aboveThis] } {
		$map_canvas raise $layer $aboveThis
	    } else {
		$map_canvas raise $layer
	    }
	    set aboveThis $layer
	}
    }


    # Update layer dialog if present

    variable ${ns}::layer_dlg
    if {[info exists layer_dlg] && [winfo exists $layer_dlg]} {
	set_layers_dlg $map
    }
}

# ::geomap::wdgeomap::get_layers --
#
#	Retrieve the map's layer list.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	Return value is the map's layer list.

proc ::geomap::wdgeomap::get_layers {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::layers

    if { ![info exists layers] } {
	set layers {}
    }
    return $layers
}

# ::geomap::wdgeomap::add_layer --
#
#	Add a set of items to the layer list.
#
# Arguments:
#	map	- map identifier.
#	layer	- name for the layer.  The map layer procedures will
#		  manage the stacking order for items with tag "$layer".
#
# Results:
#	The map's layer list is modified.  The stacking order for items
#	in the canvas might change.

proc ::geomap::wdgeomap::add_layer {map layer} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::layers

    # Skip if layer is already in layers

    if {[lsearch $layers $layer] >= 0} {
	return
    }

    # Find out if layer belongs in a hierarchy in layers.
    # If so, insert it at the end of that tree.  Otherwise,
    # insert it at the end of the layers list.

    set sub {}
    foreach elem $layer {
	lappend sub $elem
	lappend subs $sub
    }
    set long ""
    set i 0
    set i_long [llength $layers]
    foreach layer0 $layers {
	foreach sub $subs {
	    set ll [expr {[llength $sub] - 1}]
	    set sub0 [lrange $layer0 0 $ll]
	    if {$sub0 == $sub && [llength $sub0] >= [llength $long]} {
		set long $sub0
		set i_long $i
	    }
	}
	incr i
    }
    set layers [linsert $layers [expr {$i_long + 1}] $layer]
}

# ::geomap::wdgeomap::update_tree --
#
#	Update the layer hierarchy
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	The map's layer list is modified.  The stacking order for items
#	in the canvas might change.

proc ::geomap::wdgeomap::update_tree {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::layers
    variable ${ns}::items
    variable ${ns}::leaves
    variable ${ns}::parent
    variable ${ns}::children
    variable ${ns}::all_children
    variable ${ns}::leaves

    set items {}
    array unset children
    array unset all_children
    array unset leaves
    array unset parent
    foreach layer $layers {
	set ll [llength $layer]
	for {set i 0} {$i < $ll} {incr i} {
	    set item [lrange $layer 0 $i]
	    if {[lsearch $items $item] < 0} {
		lappend items $item
		set children($item) {}
		set all_children($item) {}
		set leaves($item) {}
		set parent($item) {}
	    }
	    lappend leaves($item) $layer
	    set i1 [expr {$i + 1}]
	    if {$i1 < $ll} {
		set child [lrange $layer 0 $i1]
		if {$child != ""} {
		    lappend children($item) $child
		}
		set child $item
		foreach addit [lrange $layer $i1 end] {
		    lappend child $addit
		    lappend all_children($item) $child
		}
	    }
	}
    }
    foreach item $items {
	foreach child $children($item) {
	    set parent($child) $item
	}
    }
}

# ::geomap::wdgeomap::rm_layer --
# 
#	Removes items from the layer list.
#
# Arguments:
#	map	- map identifier.
#	layer	- the layers to remove.
#
# Results:
#	The map layers list and geomap default layers list are updated.

proc ::geomap::wdgeomap::rm_layer {map layer} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::layers

    if { ![info exists layers] } {
	return
    }
    set l [lsearch $layers $layer]
    set layers [lreplace $layers $l $l]
}

# ::geomap::wdgeomap::set_layers_dlg --
#
#	This procedure creates a window in which the user can
#	change the layer order by dragging associated canvas items.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	A dialog appears with items that represent layers.
#	Map layer sequence changes as user drags items in the dialog.

proc ::geomap::wdgeomap::set_layers_dlg {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::margin 24
    variable ${ns}::layers
    variable ${ns}::items
    variable ${ns}::children
    variable ${ns}::leaves
    variable ${ns}::item_fm_id
    variable ${ns}::text_id
    variable ${ns}::box_id
    variable ${ns}::line_id
    variable ${ns}::column
    variable ${ns}::layer_dlg .[name $map]_layers
    variable ${ns}::layer_cvs ${layer_dlg}.c

    update_tree $map

    # Create a toplevel widget and canvas for the dialog.

    if { ![winfo exists $layer_dlg] } {
	toplevel $layer_dlg
	bindtags $layer_dlg [linsert [bindtags $layer_dlg] 0 wdgeomap_layer_dlg]
    } else {
	raise $layer_dlg
    }
    if { ![winfo exists $layer_cvs] } {
	::canvas $layer_cvs
    }
    set layer_close ${layer_dlg}.close 
    if { ![winfo exists $layer_close] } {
	button $layer_close -text "Close" -command "destroy $layer_dlg"
    }

    # Create or recreate items

    $layer_cvs delete layer

    set widest 0.0
    set tallest 0.0
    set c_max -1

    array unset column
    foreach item $items {
	set c [llength $item]
	set c_max [expr {($c > $c_max) ? $c : $c_max}]
	if { ![info exists column($c)] || [lsearch $column($c) $item] < 0} {
	    lappend column($c) $item
	}
	set txt [lindex $item end]
	set id [$layer_cvs create text 0 0 -text $txt \
		-fill yellow -anchor w -tags [list layer text $item]]
	set text_id($item) $id
	set item_fm_id($id) $item
	set font_desc [$layer_cvs itemcget $id -font]
	set asc [font metrics $font_desc -ascent]
	set des [font metrics $font_desc -descent]
	set h [expr {$asc + $des + 4}]
	set tallest [expr {($h > $tallest) ? $h : $tallest}]
	set w [font measure $font_desc $txt]
	set widest [expr {($w > $widest) ? $w : $widest}]
    }

    # Postion items

    set y $margin
    foreach layer $layers {
	set c [llength $layer]
	set x [expr {$margin + ($c - 1) * $widest + 3}]
	$layer_cvs coords $text_id($layer) $x [y_canvas $map $y]
	set y [expr {$y + $tallest + 3}]
    }

    set jj [array names column]
    set l [llength $jj]
    foreach j [lrange [lsort -integer -decreasing $jj] 1 [expr {$l - 1}]] {
	foreach item $column($j) {
	    if {[llength $children($item)] > 0} {
		# Position parent halfway up column of children

		set c [llength [split $item :]]
		set x [expr {$margin + ($c - 1) * $widest + 3}]
		set ids {}
		foreach child $children($item) {
		    lappend ids $text_id($child)
		}
		set bbox [eval $layer_cvs bbox $ids]
		set y1 [lindex $bbox 1]
		set y2 [lindex $bbox 3]
		set y [expr {0.5 * ($y1 + $y2) + 3}]
		$layer_cvs coords $text_id($item) $x $y

		# Draw lines from parent to children

		set p1 [$layer_cvs coords $text_id($item)]
		foreach child $children($item) {
		    set p2 [$layer_cvs coords $text_id($child)]
		    set id [$layer_cvs create line [concat $p1 $p2] \
			    -fill black -tags [list layer line $item]]
		    set parent_id $text_id($item)
		    set child_id $text_id($child)
		    set line_id(${parent_id},${child_id}) $id
		}
	    }
	}
    }

    # Give the text a black background

    foreach item $items {
	set bbox [$layer_cvs bbox $text_id($item)]
	set id [$layer_cvs create rectangle $bbox -fill black \
		-tags [list layer rect $item]]
	set box_id($item) $id
	set item_fm_id($id) $item
    }
    $layer_cvs raise layer&&text

    # Make everything visible.

    set bbox [$layer_cvs bbox layer]
    if {[llength $bbox] == 4} {
	set y1 [expr {[lindex $bbox 1]}]
	$layer_cvs move layer 0 [expr {-$y1 + $margin}]
    }
    set bbox [$layer_cvs bbox layer]
    if {[llength $bbox] == 4} {
	set x2 [expr {[lindex $bbox 2] + $margin}]
	set y2 [expr {[lindex $bbox 3] + $margin}]
	$layer_cvs configure -width $x2 -height $y2
    }
    pack $layer_cvs
    pack ${layer_dlg}.close

    # The items representing the layers can be moved up
    # and down by dragging.  At button release, the new
    # layer sequence is applied to the map.

    $layer_cvs bind layer&&(text||rect) <Button-1>		\
	    [namespace code [list set ${ns}::y0 %y]]
    $layer_cvs bind layer&&(text||rect) <B1-Motion>		\
	    [namespace code [list layers_dlg_drag $map %y]]
    $layer_cvs bind layer&&(text||rect) <ButtonRelease-1>	\
	    [namespace code [list layers_dlg_brelease $map]]
}

# ::geomap::wdgeomap::layers_dlg_drag --
#
#	This is the callback for drag events in a layer dialog.
#
# Arguments:
#	map	- map identifier.
#	y	- cursor y coordinate
#
# Results:
#	A layer item and its children are repositioned on the canvas.

proc ::geomap::wdgeomap::layers_dlg_drag {map y} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::layer_cvs
    variable ${ns}::y0
    variable ${ns}::item_fm_id
    variable ${ns}::items
    variable ${ns}::line_id
    variable ${ns}::text_id
    variable ${ns}::box_id
    variable ${ns}::children
    variable ${ns}::all_children

    set dy [expr {$y - $y0}]
    set curr_id [$layer_cvs find withtag current]
    set item $item_fm_id($curr_id)
    $layer_cvs move $text_id($item) 0 $dy
    $layer_cvs move $box_id($item) 0 $dy
    foreach child $all_children($item) {
	$layer_cvs move $text_id($child) 0 $dy
	$layer_cvs move $box_id($child) 0 $dy
    }

    # Update lines

    foreach item $items {
	if {[llength $children($item)] > 0} {
	    set p1 [$layer_cvs coords $text_id($item)]
	    foreach child $children($item) {
		set p2 [$layer_cvs coords $text_id($child)]
		set parent_id $text_id($item)
		set child_id $text_id($child)
		set id $line_id(${parent_id},${child_id})
		$layer_cvs coords $id [concat $p1 $p2]
	    }
	}
    }
    set y0 $y
}

# ::geomap::wdgeomap::layers_dlg_brelease --
#
#	This is the callback for button release events in a layer dialog.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	The layer dialog is updated.
#	Stacking order of map items assigned to layers is adjusted to
#	match the final sequence in the dialog.

proc ::geomap::wdgeomap::layers_dlg_brelease {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::layer_cvs
    variable ${ns}::item_fm_id
    variable ${ns}::parent
    variable ${ns}::children
    variable ${ns}::column
    variable ${ns}::leaves
    variable ${ns}::layers

    # At end of drag, rearrange layer order according to the vertical
    # sequence in the layer display.

    # siblings are items at the same depth and with the same parent
    # as the item that just moved.  Only the dragged item, its siblings,
    # and their associated children will move, and they will only move
    # relative to each other.  Thus, an item cannot be dragged to another
    # branch in the layer hierarchy.

    # The all_leaves list contains layers descended from the item and its
    # siblings.  It is a sublist from the layers list.  all_leaves is
    # removed from the layers list, re-ordered according to the vertical
    # sequence of siblings in the display (determined by the item_cmp proc),
    # and reinserted into the layers list at the same place.

    set curr_id [$layer_cvs find withtag current]
    set item $item_fm_id($curr_id)
    if {$parent($item) != ""} {
	set p $parent($item)
	set siblings $children($p)
    } else {
	set siblings $column(1)
    }
    set all_leaves {}
    foreach sibling $siblings {
	set all_leaves [concat $all_leaves $leaves($sibling)]
    }
    set start [lsearch $layers [lindex $all_leaves 0]]
    set len [llength $all_leaves]
    set last [expr {$start + $len - 1}]
    set layers [lreplace $layers $start $last]
    set sort_cmd [namespace code [list item_cmp $map]]
    set siblings [lsort -command $sort_cmd $siblings]
    set all_leaves {}
    foreach sibling $siblings {
	set all_leaves [concat $all_leaves $leaves($sibling)]
    }
    set layers [eval linsert \$layers $start $all_leaves]
    set_layers $map
}

# Compare two items by distance up.

proc ::geomap::wdgeomap::item_cmp {map item1 item2} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::layer_cvs
    variable ${ns}::text_id
    set y1 [y_cart $map $layer_cvs $text_id($item1)]
    set y2 [y_cart $map $layer_cvs $text_id($item2)]
    expr {$y1 < $y2 ? -1 : $y2 < $y1 ? 1 : 0}
}

# Return cartesian y (distance up from y=0) for an item.

proc ::geomap::wdgeomap::y_cart {map cvs item} {
    set y_cvs [lindex [$cvs coords $item] 1]
    return [expr {-$y_cvs}]
}

# Return canvas y from cartresian y

proc ::geomap::wdgeomap::y_canvas {map y_cart} {
    return [expr {-$y_cart}]
}

# ::geomap::wdgeomap::addmenu --
#
#	Add a menu to the wdgeomap menu bar.
#
# Arguments:
#	map	- map identifier.
#	name	- menu name.  The will be the label on the menu button
#		  and the part of the path name of the new button and
#		  menu.
#		  The name also refers to the menu later, such as when
#		  it is deleted.
#		
# Results:
#	Return value is the path name of the new menu.
#	A menu button is added to the menu bar.

proc ::geomap::wdgeomap::addmenu {map name} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::mbar
    variable ${ns}::mbutton

    if [regexp {\s} $name] {
	error "Menu name cannot contain whitespace"
    }
    set lName [string tolower $name]
    set button_path ${mbar}.$lName
    set menu_path ${button_path}.menu
    if { [winfo exists $button_path] } {
	return $menu_path
    }
    menubutton $button_path -text $name -menu $menu_path
    menu $menu_path
    pack $button_path -side left
    set mbutton($name) $button_path
    return $menu_path
}

# ::geomap::wdgeomap::getmenu --
#
#	Retrieve the path name of a menu button.
#
# Arguments:
#	map	- map identifier.
#	name	- menu name, should have been given to addmenu.
#		
# Results:
#	Return value is the path name of the menu button associated with name,
#	or "" if there is no menu by that name.

proc ::geomap::wdgeomap::getmenu {map name} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::mbutton

    if [info exists mbutton($name)] {
	return $mbutton($name)
    } else {
	return ""
    }
}

# ::geomap::wdgeomap::deletemenu --
#
#	Deletes a menu created with addmenu.
#
# Arguments:
#	map	- map identifier.
#	name	- menu name that was given to addmenu
#
# Results:
#	A menu button and its child are destroyed.

proc ::geomap::wdgeomap::deletemenu {map name} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::mbutton

    destroy $mbutton($name)
    unset mbutton($name)
}

# ::geomap::wdgeomap::ctr_reflon --
#
#	Reset the map when the mouse button is released.  It is used
#	in bind scripts for certain projections.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	Items in the map_canvas canvas are reconfigured.

proc ::geomap::wdgeomap::ctr_reflon {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::projname
    variable ${ns}::lazy
    variable ${ns}::update

    set toplevel [winfo toplevel $map_canvas]
    set oldCursor [$toplevel cget -cursor]
    $toplevel configure -cursor watch
    update

    # Save the update script so we don't run it twice
    # (once for refpoint, again for projection)

    set tmp_update $update
    set update {}

    set xCtr [expr {[winfo width $map_canvas] / 2}]
    set yCtr [expr {[winfo height $map_canvas] / 2}]
    configure $map -refpoint [xytolatlon $map $xCtr $yCtr]
    foreach item [$map_canvas find withtag geomap] {
	$map_canvas coords $item $xCtr $yCtr
    }
    if !$lazy {
	configure $map -projname $projname
    }

    # Now, run the update script and restore it.

    set update $tmp_update
    uplevel #0 $update
    $toplevel configure -cursor $oldCursor
}

# ::geomap::wdgeomap::ctr_refpt --
#
#	Sets the map and projection reference points to the given point,
#	and resets the bindings.  It is used in bind scripts for reference
#	point projections.
#
# Arguments:
#	map	- map identifier.
#	x, y	- canvas coordinates of the point which will become
#		  the new map refpoint and projection reference point.
#
# Results:
#	geomap_lnarr and geomap_place items in the canvas are
#	reconfigured.

proc ::geomap::wdgeomap::ctr_refpt {map x y} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::projname

    set toplevel [winfo toplevel $map_canvas]
    set oldCursor [$toplevel cget -cursor]
    $toplevel configure -cursor watch
    update
    configure $map -refpoint [xytolatlon $map $x $y]
    if {$projname == "PolarStereographic"} {
	configure $map -projname Stereographic
    } else {
	configure $map -projname $projname
    }
    $toplevel configure -cursor $oldCursor
}

# ::geomap::wdgeomap::setcolor --
#
#	Sets a color in the map.
#
# Arguments:
#	map	- map identifier.
#	elem		- a color element, e.g. land, rivers, places,
#			  etc. to choose the color for.
#	value		- what color to use when displaying the element.
#			  If value is "", the element is deleted.
#
# Results:
#	Map arrays are updated.  The colorscript for the element
#	is evaluated.

proc ::geomap::wdgeomap::setcolor {map elem value} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::colorval
    variable ${ns}::colorscript
    variable ${ns}::colorvals

    # Save the current color configuration in the colorvals list
    # for later undo.

    if { [info exists colorvals] } {
	set colorvals [linsert $colorvals 0 [array get colorval]]
    } else {
	set colorvals [list [array get colorval]]
    }

    # Update the colored element with user selection

    if {$value != ""} {
	set colorval($elem) $value
	if [info exists colorscript($elem)] {
	    namespace eval :: $colorscript($elem)
	} else {
	    set colorscript($elem) {}
	}
    } else {
	array unset colorval $elem
	array unset colorscript $elem
    }
}

# ::geomap::wdgeomap::setcolorscript --
#
#	Specify a script to run when a color changes.
#
# Arguments:
#	map	- map identifier.
#	elem	- a color element, e.g. land, rivers, places,
#			  etc. to choose the color for.
#	script	- script to run when the color of elem changes.
#
# Results:

proc ::geomap::wdgeomap::setcolorscript {map elem script} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::colorscript
    set colorscript($elem) $script
}

# ::geomap::wdgeomap::getcolor --
#
#	Get the color specified for an element in a map.
#
# Arguments:
#	map	- map identifier.
#	elem		- a color element, e.g. land, rivers, places,
#			  etc. to choose the color for.
#
# Results:
#	Returns the color to use when displaying the element in the map,
#	or "" if their is no color for the element.

proc ::geomap::wdgeomap::getcolor {map elem} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::colorval
    if [info exists colorval($elem)] {
	return $colorval($elem)
    } else {
	return ""
    }
}

# ::geomap::wdgeomap::choose_color --
#
#	Ask user to choose a new color in a dialog.  This is the callback
#	for buttons created by the File->Colors menu.  When activated by a
#	button press, it prompts the user for a color to use in the map.
#
# Arguments:
#	map	- map identifier.
#	elem	- a color element, e.g. land, rivers, places,
#		  etc. to choose the color for.
#
# Results:
#	Return value should be ignored.
#	colorval($elem) is set to the value chosen by the
#	user.  The colorval array is dumped to the colorvals list
#	for later retrieval with the Undo and Redo buttons.

proc ::geomap::wdgeomap::choose_color {map elem} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::colorval
    variable ${ns}::colorscript
    variable ${ns}::colorvals
    variable ${ns}::color_dlg

    set new [tk_chooseColor -title "$map $elem" -initialcolor $colorval($elem)]
    if {[string length $new] > 0} {
	setcolor $map $elem $new
	set btn ${color_dlg}.[string tolower $elem]
	$btn configure -background $colorval($elem)
	${color_dlg}.f.undo configure -state normal
    }
}

# ::geomap::wdgeomap::load_colors
#
#	Modify map colors as directed from a file.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	User selects a file from a dialog.  Map colors change according
#	to file contents.

proc ::geomap::wdgeomap::load_colors {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::colorval
    variable ${ns}::colorscript
    variable ${ns}::colorvals
    variable ${ns}::color_dlg

    set f [tk_getOpenFile]
    if {$f != ""} {

	# Search the input file for a line starting with "colors:"
	# and set the colorval array from its contents.

	set in [open $f]
	while {[gets $in line] >= 0} {
	    if [regexp {^colors:[ 	]*(.*)} $line m arr] {

		# Save the current color configuration in the
		# colorvals list for later undo.

		if { [info exists colorvals] } {
		    set colorvals [linsert $colorvals 0 [array get colorval]]
		} else {
		    set colorvals [list [array get colorval]]
		}
		${color_dlg}.f.undo configure -state normal

		# Set the new color configuration

		array set colorval $arr
		foreach elem [array names colorval] {
		    set btn ${color_dlg}.[string tolower $elem]
		    $btn configure -background $colorval($elem)
		    namespace eval :: $colorscript($elem)
		}
	    }
	}
	close $in
    }
}

# ::geomap::wdgeomap::save_colors
#
#	Save map colors into a file readable with the load_colors procedure.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	User selects a file from a dialog.  Map colors change according
#	to file contents.

proc ::geomap::wdgeomap::save_colors {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::colorval

    set f [tk_getSaveFile]
    if {$f != ""} {
	set out [open $f w]
	puts $out "
# Map colors --
# The list following \"colors:\" below specifies map color scheme.
# The list is a set of key value pairs.  The keys identify items to be
# colored, such as \"water\" or \"land.\"  Each key is followed by a color
# value, such as \"Blue4\" or \"wheat\" indicating the color to use for the
# corresponding item.
	"
	puts $out "colors: [array get colorval]"
	close $out
    }
}

# ::geomap::wdgeomap::undo_colors
#
#	Revert to previous color scheme.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	Colors in the map revert to the previous scheme.
#	colorval array and colorvals list are updated.
#	Scripts from the colorscript array are called.

proc ::geomap::wdgeomap::undo_colors {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::colorval
    variable ${ns}::colorscript
    variable ${ns}::colorvals
    variable ${ns}::color_dlg

    array set colorval [lindex $colorvals 0]
    set colorvals [lrange $colorvals 1 end]
    if {[llength $colorvals] == 0} {
	${color_dlg}.f.undo configure -state disabled
    }
    foreach elem [array names colorval] {
	set btn ${color_dlg}.[string tolower $elem]
	$btn configure -background $colorval($elem)
	if { [info exists colorscript($elem)] } {
	    namespace eval :: $colorscript($elem)
	}
    }
}

# ::geomap::wdgeomap::set_colors_dlg
#
#	Creates a dialog in which the user can choose new colors for
#	certain map items.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	A dialog box appears that enables modification of certain canvas
#	items.  The dialog does not block the application.  It exists
#	until the user destroys it.

proc ::geomap::wdgeomap::set_colors_dlg {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::colorval
    variable ${ns}::colorscript
    variable ${ns}::colorvals

    variable ${ns}::color_dlg .[name $map]_colors
    if { [winfo exists $color_dlg] } {
	raise $color_dlg
	return
    }

    # Create a dialog box

    toplevel $color_dlg
    bindtags $color_dlg [linsert [bindtags $color_dlg] 0 wdgeomap_color_dlg]

    # For each color element in the map, create a
    # button to modify the color.

    foreach elem [lsort [array names colorval]] {
	set btn $color_dlg.[string tolower $elem]
	set cmd [namespace code [list choose_color $map $elem]]
	button $btn -text $elem -background $colorval($elem) -command $cmd
	pack $btn -fill x
    }

    # Button row for action buttons

    frame ${color_dlg}.f
    button ${color_dlg}.f.load -text "Load" \
	    -command [namespace code [list load_colors $map]]
    button ${color_dlg}.f.save -text "Save" \
	    -command [namespace code [list save_colors $map]]

    # "Undo" button moves one element down in the colorvals list and
    # displays the color scheme stored in that element in the dialog.
    # If we end up at start of colorvals list, dim the "Undo" button.

    if {[info exists colorvals] && [llength $colorvals] > 0} {
	set st "normal"
    } else {
	set st "disabled"
    }
    button ${color_dlg}.f.undo -text "Undo" -state $st \
	    -command [namespace code [list undo_colors $map]]

    # Close button

    button ${color_dlg}.f.close -text "Close" -command "destroy $color_dlg"

    pack ${color_dlg}.f.undo ${color_dlg}.f.save ${color_dlg}.f.load \
	    ${color_dlg}.f.close -side left
    pack ${color_dlg}.f
}

# postscript --
#
#	Generates postscript for the canvas based on user input from
#	a dialog box.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	A dialog box appears in which the user gives the name of a
#	postscript file.  The dialog does not block the application,
#	and it remains until the user destroys it.

proc ::geomap::wdgeomap::postscript {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_canvas
    variable ${ns}::postscript_file
    variable ${ns}::color_mode
    variable ${ns}::ps_dlg

    set ps_dlg .[name $map]_postscript
    if [winfo exists $ps_dlg] {
	raise $ps_dlg
	return
    }

    # Create a dialog box prompting user for a name for a postscript file.

    toplevel $ps_dlg
    bindtags $ps_dlg [linsert [bindtags $ps_dlg] 0 wdgeomap_ps_dlg]
    set x [expr {[winfo x $map_canvas] + 200}]
    set y [expr {[winfo y $map_canvas] + 200}]
    wm geometry $ps_dlg +$x+$y
    label ${ps_dlg}.msg -text "Send postscript to file"
    entry ${ps_dlg}.e
    if ![info exists postscript_file] {
	set postscript_file "c.ps"
    }
    ${ps_dlg}.e insert 0 $postscript_file
  
    # Radio buttons for colormode

    if { ![info exists color_mode] } {
	set color_mode gray
    }
    frame ${ps_dlg}.r
    radiobutton ${ps_dlg}.r.color -text color \
	    -variable ${ns}::color_mode -value color
    radiobutton ${ps_dlg}.r.gray -text gray \
	    -variable ${ns}::color_mode -value gray
    radiobutton ${ps_dlg}.r.mono -text mono \
	    -variable ${ns}::color_mode -value mono
    pack ${ps_dlg}.r.color ${ps_dlg}.r.gray ${ps_dlg}.r.mono \
	    -side left -expand 1 -fill x
  
    # OK and Cancel

    frame ${ps_dlg}.b
    set callback [list namespace eval $ns {
	    set postscript_file [${ps_dlg}.e get]
	    if {[string length $postscript_file] == 0} {
		tk_messageBox -title "Error" -type ok \
			-message "Must have postscript file name"
		destroy $ps_dlg
		return
	    }
	    $map_canvas postscript -colormode $color_mode -file $postscript_file
	    destroy $ps_dlg
	    tk_messageBox -message "Created $postscript_file"
	}]
    bind ${ps_dlg}.e <Return> $callback
    button ${ps_dlg}.b.ok -text OK -command $callback
    button ${ps_dlg}.b.cancel -text Cancel -command [list destroy $ps_dlg]

    pack ${ps_dlg}.b.ok ${ps_dlg}.b.cancel -side left
    pack ${ps_dlg}.msg ${ps_dlg}.e ${ps_dlg}.r ${ps_dlg}.b
}

# ::geomap::wdgeomap::delete --
#
#	Delete the map and associated commands and variables.
#
# Arguments:
#	map	- map identifier.
#
# Results:
#	The namespace for the map, and its associated commands and
#	variables are deleted.

# This procedure should be called to delete the map.

proc ::geomap::wdgeomap::delete {map} {
    variable map_ns
    set ns $map_ns($map)
    variable ${ns}::map_frame
    variable ${ns}::projId

    destroy $map_frame
    rename $projId {}
    namespace delete $ns
}

# ::geomap::wdgeomap::set_motion_bindings --
#
#	This procedure sets the bindings that allow the user to move a map
#	and adjust its reference point.
#
# Arguments:
#	modifier	- event modifier, e.g. "Control" or "Shift" to require
#			  for mouse actions intended to move the map.
#	button		- mouse button to push to move the map.
#
# Results:
#	Bindings to widgets with the geomap_motion tag are modified.

proc ::geomap::wdgeomap::set_motion_bindings {modifier button} {
    set mods [list {} Alt Control Shift]
    if {[lsearch -exact $mods $modifier] < 0} {
	error "Modifier must be one of $mods"
    }
    if {$modifier != ""} {
	set modifier ${modifier}-
    }
    if ![string is integer $button] {
	error "Button identifier must be an integer"
    }

    foreach binding [bind geomap_motion] {
	bind geomap_motion $binding {}
    }

    # Update map bindings.

    set double "<${modifier}Double-Button-$button>"
    set press "<${modifier}Button-$button>"
    set motion "<${modifier}Button$button-Motion>"
    set release "<${modifier}ButtonRelease-$button>"

    # If map projection uses reference longitude, dragging moves the map.

    bind geomap_motion $press [namespace code {
	set toplev [winfo toplevel %W]
	set map $map_tl($toplev)
	set proj_info [::geomap::proj_info [cget $map -projname]]
	set ref [lindex $proj_info 0]
	if {$ref == "longitude"} {
	    set x00 %x
	    set y00 %y
	    set x0 %x
	    set y0 %y
	}
    }]
    bind geomap_motion $motion [namespace code {
	set toplev [winfo toplevel %W]
	set map $map_tl($toplev)
	set proj_info [::geomap::proj_info [cget $map -projname]]
	set ref [lindex $proj_info 0]
	if {$ref == "longitude"} {
	    set map_canvas [map_canvas $map]
	    if {[info exists x0] && [info exists y0]} {
		$map_canvas move geomap [expr {%x - $x0}] [expr {%y - $y0}]
	    }
	    set x0 %x
	    set y0 %y
	}
    }]
    bind geomap_motion $release [namespace code {
	set toplev [winfo toplevel %W]
	set map $map_tl($toplev)
	set proj_info [::geomap::proj_info [cget $map -projname]]
	set ref [lindex $proj_info 0]
	if {$ref == "longitude"} {
	    if {[info exists x00] && [info exists y00]} {
		if {%x != $x00 && %y != $y00} {
		    ctr_reflon $map
		}
		unset -nocomplain x00
		unset -nocomplain y00
	    }
	    unset -nocomplain x0
	    unset -nocomplain y0
	}
    }]

    # If map projection uses a reference point, double clicking moves
    # the map.

    bind geomap_motion $double [namespace code {
	set toplev [winfo toplevel %W]
	set map $map_tl($toplev)
	set proj_info [::geomap::proj_info [cget $map -projname]]
	set ref [lindex $proj_info 0]
	if {$ref == "point"} {
	    ctr_refpt $map %x %y
	}
    }]
}

# ::geomap::wdgeomap::get_motion_bindings --
#
#	This procedure sets the bindings that allow the user to move a map
#	and adjust its reference point.
#
# Arguments:
#	None.
#
# Results:
#	Return value is a list of bindings for the geomap_motion tag.

proc ::geomap::wdgeomap::get_motion_bindings {} {
    return [bind geomap_motion]
}
